var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
import { Connection } from "./Connection";
import { ConnectionNotFoundError } from "./error/ConnectionNotFoundError";
import { MissingDriverError } from "./error/MissingDriverError";
import { AlreadyHasActiveConnectionError } from "./error/AlreadyHasActiveConnectionError";
import { Logger } from "../logger/Logger";
import { CannotDetermineConnectionOptionsError } from "./error/CannotDetermineConnectionOptionsError";
import { WebSqlDriver } from "../driver/websql/WebSqlDriver";
import { IonicSQLiteDriver } from "../driver/ionic-sqlite/IonicSQLiteDriver";
/**
 * ConnectionManager is used to store and manage all these different connections.
 * It also provides useful factory methods to simplify connection creation.
 */
var ConnectionManager = (function () {
    function ConnectionManager() {
        // -------------------------------------------------------------------------
        // Protected Properties
        // -------------------------------------------------------------------------
        /**
         * List of connections registered in this connection manager.
         */
        this.connections = [];
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Checks if connection with the given name exist in the manager.
     */
    ConnectionManager.prototype.has = function (name) {
        return !!this.connections.find(function (connection) { return connection.name === name; });
    };
    /**
     * Gets registered connection with the given name.
     * If connection name is not given then it will get a default connection.
     * Throws exception if connection with the given name was not found.
     */
    ConnectionManager.prototype.get = function (name) {
        if (name === void 0) { name = "default"; }
        var connection = this.connections.find(function (connection) { return connection.name === name; });
        if (!connection)
            throw new ConnectionNotFoundError(name);
        return connection;
    };
    /**
     * Creates a new connection based on the given connection options and registers it in the manager.
     * You need to manually call #connect method to establish connection.
     * Note that dropSchemaOnConnection and autoSchemaSync options of a ConnectionOptions will not work there - use
     * createAndConnect method to use them.
     */
    ConnectionManager.prototype.create = function (options) {
        var logger = new Logger(options.logging || {});
        var driver = this.createDriver(options.driver, logger);
        var connection = this.createConnection(options.name || "default", driver, logger);
        // import entity schemas
        if (options.entitySchemas) {
            var _a = this.splitStringsAndClasses(options.entitySchemas), directories = _a[0], classes = _a[1];
            connection
                .importEntitySchemas(classes);
            // .importEntitySchemaFromDirectories(directories);
        }
        // import entities
        if (options.entities) {
            var _b = this.splitStringsAndClasses(options.entities), directories = _b[0], classes = _b[1];
            connection
                .importEntities(classes);
            // .importEntitiesFromDirectories(directories);
        }
        // import subscriber
        if (options.subscribers) {
            var _c = this.splitStringsAndClasses(options.subscribers), directories = _c[0], classes = _c[1];
            connection
                .importSubscribers(classes);
            // .importSubscribersFromDirectories(directories);
        }
        // import naming strategies
        if (options.namingStrategies) {
            var _d = this.splitStringsAndClasses(options.namingStrategies), directories = _d[0], classes = _d[1];
            connection
                .importNamingStrategies(classes);
            // .importNamingStrategiesFromDirectories(directories);
        }
        // set naming strategy to be used for this connection
        if (options.usedNamingStrategy)
            connection.useNamingStrategy(options.usedNamingStrategy);
        return connection;
    };
    /**
     * Creates a new connection and registers it in the manager.
     *
     * If connection options were not specified, then it will try to create connection automatically.
     *
     * First, it will try to find a "default" configuration from ormconfig.json.
     * You can also specify a connection name to use from ormconfig.json,
     * and you even can specify a path to custom ormconfig.json file.
     *
     * In the case if options were not specified, and ormconfig.json file also wasn't found,
     * it will try to create connection from environment variables.
     * There are several environment variables you can set:
     *
     * - TYPEORM_DRIVER_TYPE - driver type. Can be "mysql", "mysql2", "postgres", "mariadb", "websql", "oracle" or "mssql".
     * - TYPEORM_URL - database connection url. Should be a string.
     * - TYPEORM_HOST - database host. Should be a string.
     * - TYPEORM_PORT - database access port. Should be a number.
     * - TYPEORM_USERNAME - database username. Should be a string.
     * - TYPEORM_PASSWORD - database user's password. Should be a string.
     * - TYPEORM_SID - database's SID. Used only for oracle databases. Should be a string.
     * - TYPEORM_STORAGE - database's storage url. Used only for websql databases. Should be a string.
     * - TYPEORM_USE_POOL - indicates if connection pooling should be enabled. By default its enabled. Should be boolean-like value.
     * - TYPEORM_DRIVER_EXTRA - extra options to be passed to the driver. Should be a serialized json string of options.
     * - TYPEORM_AUTO_SCHEMA_SYNC - indicates if automatic schema synchronization will be performed on each application run. Should be boolean-like value.
     * - TYPEORM_ENTITIES - list of directories containing entities to load. Should be string - directory names (can be patterns) split by a comma.
     * - TYPEORM_SUBSCRIBERS - list of directories containing subscribers to load. Should be string - directory names (can be patterns) split by a comma.
     * - TYPEORM_ENTITY_SCHEMAS - list of directories containing entity schemas to load. Should be string - directory names (can be patterns) split by a comma.
     * - TYPEORM_NAMING_STRATEGIES - list of directories containing custom naming strategies to load. Should be string - directory names (can be patterns) split by a comma.
     * - TYPEORM_LOGGING_QUERIES - indicates if each executed query must be logged. Should be boolean-like value.
     * - TYPEORM_LOGGING_FAILED_QUERIES - indicates if logger should log failed query's error. Should be boolean-like value.
     * - TYPEORM_LOGGING_ONLY_FAILED_QUERIES - indicates if only failed queries must be logged. Should be boolean-like value.
     *
     * TYPEORM_DRIVER_TYPE variable is required. Depend on the driver type some other variables may be required too.
     */
    // async createAndConnect(): Promise<Connection>;
    /**
     * Creates connection from the given connection options and registers it in the manager.
     */
    /**
     * Creates connection and and registers it in the manager.
     */
    ConnectionManager.prototype.createAndConnect = function (connectionOptions) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                // if connection options are given, then create connection from them
                if (connectionOptions && connectionOptions instanceof Object) {
                    return [2 /*return*/, this.createAndConnectByConnectionOptions(connectionOptions)];
                }
                // if connection name is specified then explicitly try to load connection options from it
                // if (typeof optionsOrConnectionNameFromConfig === "string")
                //     return this.createFromConfigAndConnect(optionsOrConnectionNameFromConfig || "default", ormConfigPath);
                // if driver type is set in environment variables then try to create connection from env variables
                // if (this.hasDefaultConfigurationInEnvironmentVariables())
                //     return this.createFromEnvAndConnect();
                throw new CannotDetermineConnectionOptionsError();
            });
        });
    };
    /**
     * Creates connections and and registers them in the manager.
     */
    ConnectionManager.prototype.createAndConnectToAll = function (optionsOrOrmConfigFilePath) {
        return __awaiter(this, void 0, void 0, function () {
            var _this = this;
            return __generator(this, function (_a) {
                // if connection options are given, then create connection from them
                if (optionsOrOrmConfigFilePath && optionsOrOrmConfigFilePath instanceof Array)
                    return [2 /*return*/, Promise.all(optionsOrOrmConfigFilePath.map(function (options) {
                            return _this.createAndConnectByConnectionOptions(options);
                        }))];
                // if connection name is specified then explicitly try to load connection options from it
                // if (typeof optionsOrOrmConfigFilePath === "string")
                //     return this.createFromConfigAndConnectToAll(optionsOrOrmConfigFilePath as string);
                // if driver type is set in environment variables then try to create connection from env variables
                // if (this.hasDefaultConfigurationInEnvironmentVariables())
                //     return [await this.createFromEnvAndConnect()];
                throw new CannotDetermineConnectionOptionsError();
            });
        });
    };
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    /**
     * Checks if environment variables contains connection options.
     */
    /*protected hasDefaultConfigurationInEnvironmentVariables(): boolean {
        return !!process.env.TYPEORM_DRIVER_TYPE;
    }*/
    /**
     * Allows to quickly create a connection based on the environment variable values.
     */
    /*protected async createFromEnvAndConnect(): Promise<Connection> {
        return this.createAndConnectByConnectionOptions({
            driver: {
                type: process.env.TYPEORM_DRIVER_TYPE,
                url: process.env.TYPEORM_URL,
                host: process.env.TYPEORM_HOST,
                port: process.env.TYPEORM_PORT,
                username: process.env.TYPEORM_USERNAME,
                password: process.env.TYPEORM_PASSWORD,
                database: process.env.TYPEORM_DATABASE,
                sid: process.env.TYPEORM_SID,
                storage: process.env.TYPEORM_STORAGE,
                usePool: process.env.TYPEORM_USE_POOL !== undefined ? OrmUtils.toBoolean(process.env.TYPEORM_USE_POOL) : undefined, // special check for defined is required here
                extra: process.env.TYPEORM_DRIVER_EXTRA ? JSON.parse(process.env.TYPEORM_DRIVER_EXTRA) : undefined
            },
            autoSchemaSync: OrmUtils.toBoolean(process.env.TYPEORM_AUTO_SCHEMA_SYNC),
            entities: process.env.TYPEORM_ENTITIES ? process.env.TYPEORM_ENTITIES.split(",") : [],
            subscribers: process.env.TYPEORM_SUBSCRIBERS ? process.env.TYPEORM_SUBSCRIBERS.split(",") : [],
            entitySchemas: process.env.TYPEORM_ENTITY_SCHEMAS ? process.env.TYPEORM_ENTITY_SCHEMAS.split(",") : [],
            namingStrategies: process.env.TYPEORM_NAMING_STRATEGIES ? process.env.TYPEORM_NAMING_STRATEGIES.split(",") : [],
            usedNamingStrategy: process.env.TYPEORM_USED_NAMING_STRATEGY,
            logging: {
                logQueries: OrmUtils.toBoolean(process.env.TYPEORM_LOGGING_QUERIES),
                logFailedQueryError: OrmUtils.toBoolean(process.env.TYPEORM_LOGGING_FAILED_QUERIES),
                logOnlyFailedQueries: OrmUtils.toBoolean(process.env.TYPEORM_LOGGING_ONLY_FAILED_QUERIES),
            }
        });
    }*/
    /**
     * Creates a new connection based on the connection options from "ormconfig.json"
     * and registers a new connection in the manager.
     * Optionally you can specify a path to the json configuration.
     * If path is not given, then ormconfig.json file will be searched near node_modules directory.
     */
    /* protected async createFromConfigAndConnectToAll(path?: string): Promise<Connection[]> {
         const optionsArray: ConnectionOptions[] = require(path || (require("app-root-path").path + "/ormconfig.json"));
         if (!optionsArray)
             throw new Error(`Configuration ${path || "ormconfig.json"} was not found. Add connection configuration inside ormconfig.json file.`);
 
         const promises = optionsArray
             .filter(options => !options.environment || options.environment === process.env.NODE_ENV) // skip connection creation if environment is set in the options, and its not equal to the value in the NODE_ENV variable
             .map(options => this.createAndConnectByConnectionOptions(options));
 
         return Promise.all(promises);
     }*/
    /**
     * Creates a new connection based on the connection options from "ormconfig.json"
     * and registers a new connection in the manager.
     * Optionally you can specify a path to the json configuration.
     * If path is not given, then ormconfig.json file will be searched near node_modules directory.
     */
    /*protected async createFromConfigAndConnect(connectionName: string, path?: string): Promise<Connection> {
        const optionsArray: ConnectionOptions[] = require(path || (require("app-root-path").path + "/ormconfig.json"));
        if (!optionsArray)
            throw new Error(`Configuration ${path || "ormconfig.json"} was not found. Add connection configuration inside ormconfig.json file.`);

        const environmentLessOptions = optionsArray.filter(options => (options.name || "default") === connectionName);
        const options = environmentLessOptions.filter(options => !options.environment || options.environment === process.env.NODE_ENV); // skip connection creation if environment is set in the options, and its not equal to the value in the NODE_ENV variable

        if (!options.length)
            throw new Error(`Connection "${connectionName}" ${process.env.NODE_ENV ? "for the environment " + process.env.NODE_ENV + " " : ""}was not found in the json configuration file.` +
                (environmentLessOptions.length ? ` However there are such configurations for other environments: ${environmentLessOptions.map(options => options.environment).join(", ")}.` : ""));

        return this.createAndConnectByConnectionOptions(options[0]);
    }*/
    /**
     * Creates a new connection based on the given connection options and registers a new connection in the manager.
     */
    ConnectionManager.prototype.createAndConnectByConnectionOptions = function (options) {
        return __awaiter(this, void 0, void 0, function () {
            var connection;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        connection = this.create(options);
                        // connect to the database
                        return [4 /*yield*/, connection.connect()];
                    case 1:
                        // connect to the database
                        _a.sent();
                        if (!options.dropSchemaOnConnection /* && !process.env.SKIP_SCHEMA_CREATION*/) return [3 /*break*/, 3]; /* && !process.env.SKIP_SCHEMA_CREATION*/
                        return [4 /*yield*/, connection.dropDatabase()];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        if (!options.autoSchemaSync /* && !process.env.SKIP_SCHEMA_CREATION*/) return [3 /*break*/, 5]; /* && !process.env.SKIP_SCHEMA_CREATION*/
                        return [4 /*yield*/, connection.syncSchema()];
                    case 4:
                        _a.sent();
                        _a.label = 5;
                    case 5: return [2 /*return*/, connection];
                }
            });
        });
    };
    /**
     * Splits given array of mixed strings and / or functions into two separate array of string and array of functions.
     */
    ConnectionManager.prototype.splitStringsAndClasses = function (strAndClses) {
        return [
            strAndClses.filter(function (str) { return typeof str === "string"; }),
            strAndClses.filter(function (cls) { return typeof cls !== "string"; }),
        ];
    };
    /**
     * Creates a new driver based on the given driver type and options.
     */
    ConnectionManager.prototype.createDriver = function (options, logger) {
        switch (options.type) {
            case "ionic-sqlite":
                return new IonicSQLiteDriver(options, logger);
            case "websql":
                return new WebSqlDriver(options, logger);
            default:
                throw new MissingDriverError(options.type);
        }
    };
    /**
     * Creates a new connection and registers it in the connection manager.
     */
    ConnectionManager.prototype.createConnection = function (name, driver, logger) {
        var existConnection = this.connections.find(function (connection) { return connection.name === name; });
        if (existConnection) {
            if (existConnection.isConnected)
                throw new AlreadyHasActiveConnectionError(name);
            this.connections.splice(this.connections.indexOf(existConnection), 1);
        }
        var connection = new Connection(name, driver, logger);
        this.connections.push(connection);
        return connection;
    };
    return ConnectionManager;
}());
export { ConnectionManager };
//# sourceMappingURL=ConnectionManager.js.map